"""
"""


#[
from __future__ import annotations

from typing import (Any, Iterable, )
from numbers import (Real, )
import functools as _ft
import dataclasses as _dc
import numpy as _np
import scipy as _sp

from .. import has_variants as _has_variants
from .. import quantities as _quantities
from .. import pages as _pages
from ..dataslates.main import (Dataslate, )
from ..databoxes.main import (Databox, )
from .solutions import (right_div, left_div, )
from ..dates import (Dater, )

from . import initializers as _ford_initializers
from . import simulators as _ford_simulators
#]


simulate_anticipated_shocks \
    = _ford_simulators.simulate_triangular_anticipated_shocks


class KalmanOutputData:
    """
......................................................................

==Output data from Kalman filter==


......................................................................
    """
    #[

    __slots__ = (
        "predict_mse_measurement",
        "predict_med",
        "predict_std",
        "predict_err",
        "update_med",
        "update_std",
        "smooth_med",
        "smooth_std",
    )

    def __init__(
        self: Self,
        output_store: _OutputStore,
    ) -> None:
        """
        """
        for attr_name in self.__slots__:
            attr = getattr(output_store, attr_name, )
            if hasattr(attr, "to_output_arg", ):
                attr = attr.to_output_arg()
            setattr(self, attr_name, attr, )

    def __repr__(self, /, ) -> str:
        """
        """
        slots_to_repr = (s for s in self.__slots__ if getattr(self, s) is not None)
        return f"KalmanOutputData({', '.join(slots_to_repr)})"

    #]


class _Needs:
    """
    """
    #[

    return_predict = True
    return_update = True
    return_smooth = True
    rescale_variance = False

    def __init__(self: Self, **kwargs, ) -> None:
        """
        """
        for k, v in kwargs.items():
            if not hasattr(self, k) or getattr(self, k) is None:
                continue
            setattr(self, k, v)

    #]


class _LogDataslate:
    """
    """
    #[

    @classmethod
    def from_dataslate(
        klass,
        input_ds: Dataslate,
        num_variants: int,
        name_to_log_name: dict[str, str],
        /,
    ) -> Self:
        """
        """
        self = klass()
        self._dataslate = Dataslate.nan_from_template(input_ds, num_variants=num_variants, )
        self._prepare_log_names(name_to_log_name, )
        return self

    @classmethod
    def from_names_dates(
        klass,
        measurement_names: Iterable[str],
        dates: Iterable[Dater],
        num_variants: int,
        name_to_log_name: dict[str, str],
    ) -> Self:
        """
        """
        self = klass()
        self._dataslate = Dataslate.nan_from_names_periods(
            measurement_names, dates,
            num_variants=num_variants,
        )
        self._prepare_log_names(name_to_log_name, )
        return self

    def _prepare_log_names(
        self,
        name_to_log_name: dict[str, str],
        /,
    ) -> None:
        """
        """
        self._dataslate.rename(name_to_log_name, )
        output_names = self._dataslate.output_names
        self._name_to_log_name = {
            name: log_name
            for name, log_name in name_to_log_name.items()
            if log_name in output_names
        }

    def extend(
        self,
        other: Self,
        /,
    ) -> None:
        """
        """
        self._dataslate.extend(other._dataslate, )

    #]


class _MedLogDataslate(_LogDataslate, ):
    """
    """
    #[

    def store(
        self: Self,
        med_array: _np.ndarray,
        lhs_indexes: tuple,
        rhs_indexes=None,
        transform=None,
    ) -> None:
        """
        """
        rhs_indexes = rhs_indexes if rhs_indexes is not None else slice(None, )
        if transform is not None:
            med_array = transform[rhs_indexes, :] @ med_array
        else:
            med_array = med_array[rhs_indexes]
        self._dataslate._variants[0].data[lhs_indexes] = med_array

    def to_output_arg(
        self: Self,
        /,
    ) -> Databox:
        """
        """
        db = self._dataslate.to_databox()
        for name, log_name in self._name_to_log_name.items():
            db[name] = db[log_name].copy()
            db[name].exp()
        return db

    #]


class _StdLogDataslate(_LogDataslate, ):
    """
    """
    #[

    def store(
        self: Self,
        mse_array: _np.ndarray,
        lhs_indexes: tuple,
        rhs_indexes=None,
        transform=None,
    ) -> None:
        rhs_indexes = rhs_indexes if rhs_indexes is not None else slice(None, )
        if transform is not None:
            transform_rhs_indexes = transform[rhs_indexes, :]
            mse_array = transform_rhs_indexes @ mse_array @ transform_rhs_indexes.T
        else:
            mse_array = mse_array[rhs_indexes, ...][..., rhs_indexes]
        self._dataslate._variants[0].data[lhs_indexes] = _std_from_mse(mse_array, )

    def to_output_arg(
        self: Self,
        /,
    ) -> Databox:
        """
        """
        return self._dataslate.to_databox()

    #]


class _OutputStore:
    """
    """
    #[

    __slots__ = (
        "predict_med",
        "predict_std",
        "predict_mse_measurement",

        "update_med",
        "update_std",
        "predict_err",

        "smooth_med",
        "smooth_std",
    )

    def __init__(
        self,
        input_ds: Dataslate,
        num_variants: int,
        measurement_names: Iterable[str],
        needs: _Needs,
        name_to_log_name: dict[str, str] | None = None,
    ) -> Self:
        """
        """
        _med_constructor = _ft.partial(
            _MedLogDataslate.from_dataslate,
            input_ds, num_variants, name_to_log_name,
        )
        _std_constructor = _ft.partial(
            _StdLogDataslate.from_dataslate,
            input_ds, num_variants, name_to_log_name,
        )
        if needs.return_predict:
            self.predict_med = _med_constructor()
            self.predict_std = _std_constructor()
            self.predict_mse_measurement = [ () for _ in range(num_variants, ) ]
        if needs.return_update:
            self.update_med = _med_constructor()
            self.update_std = _std_constructor()
            self.predict_err = _MedLogDataslate.from_names_dates(
                measurement_names,
                input_ds.periods,
                num_variants,
                name_to_log_name,
            )
        if needs.return_smooth:
            self.smooth_med = _med_constructor()
            self.smooth_std = _std_constructor()

    def extend(self, other: Self, /, ) -> None:
        """
        """
        for s in self.__slots__:
            if getattr(self, s) is not None:
                getattr(self, s).extend(getattr(other, s), )

    def rescale_stds(self, var_scale, /, ) -> None:
        """
        """
        if var_scale == 1 or var_scale is None:
            return
        std_scale = _sqrt_positive(self.var_scale)
        for attr_name in ("predict_std", "update_std", "smooth_std", ):
            attr = getattr(self, attr_name, )
            if attr is None:
                continue
            attr._variants = [ v.data * std_scale for v in attr._variants ]

    def create_output_data(self: Self, /, ) -> KalmanOutputData:
        """
        """
        return KalmanOutputData(self, )

    #]


class _Cache:
    """
    """
    #[

    LOG_2_PI = _np.log(2 * _np.pi)

    __slots_to_input__ = (
        "model_v",
        "input_ds_v",
        "solution",
        "needs",
        "curr_qids",
        "curr_pos",
        "y_qids",
        "u_qids",
        "v_qids",
        "w_qids",
        "std_u_qids",
        "std_w_qids",
        "logly_within_y",
    )

    __slots_to_preallocate__ = (
        "all_num_obs",
        "all_pex_invFx_pex",
        "all_invFx_pex",
        "all_log_det_Fx",
        "all_L",
        "all_a0",
        "all_P0",
        "all_ZaxT_invFx_Zax",
        "all_ZaxT_invFx",
        "all_G",
        "all_inx_y",
        "all_pex",
        "all_invFx",
        "all_a1",
        "all_P1",
        "all_cov_u",
        "all_cov_w",
    )

    __slots_to_create__ = (
        "num_periods",
        "base_columns",
        "nonbase_columns",
        "sum_num_obs",
        "sum_log_det_Fx",
        "sum_pex_invFx_pex",
        "y1_array",
        "u0_array",
        "v0_array",
        "w0_array",
        "all_v_impact",
        "std_u_array",
        "std_w_array",
        "var_scale",
        "neg_log_likelihood",
        "neg_log_likelihood_contributions",
        "diffuse_factor",
    )

    __slots__ = __slots_to_input__ + __slots_to_preallocate__ + __slots_to_create__

    def __init__(self: Self, **kwargs, ) -> None:
        """
        """
        for n in self.__slots_to_input__:
            setattr(self, n, kwargs[n], )
        #
        self.num_periods = self.input_ds_v.num_periods
        self.base_columns = self.input_ds_v.base_columns
        self.nonbase_columns = self.input_ds_v.nonbase_columns
        #
        # Preallocate time-varying arrayas
        for n in self.__slots_to_preallocate__:
            setattr(self, n, [None] * self.num_periods, )
        #
        # Extract arrays from input dataslate
        input_data_array = self.input_ds_v.get_data_variant(0, )
        self.y1_array = input_data_array[self.y_qids, :]
        self.u0_array = input_data_array[self.u_qids, :]
        self.v0_array = input_data_array[self.v_qids, :]
        self.w0_array = input_data_array[self.w_qids, :]
        self.std_u_array = input_data_array[self.std_u_qids, :]
        self.std_w_array = input_data_array[self.std_w_qids, :]
        if self.logly_within_y:
            self.y1_array[self.logly_within_y, :] = _np.log(self.y1_array[self.logly_within_y, :])
        #
        # Detect observations available in each period
        self.all_inx_y = [
            _np.isfinite(column_data, )
            for column_data in self.y1_array.T
        ]
        # Simulate impact of anticipated shocks
        self.all_v_impact = simulate_anticipated_shocks(
            self.model_v,
            self.input_ds_v,
        )

    def calculate_likelihood(self: Self, /, ) -> None:
        """
        """
        self.sum_num_obs = sum(i for i in self.all_num_obs if i is not None)
        self.sum_log_det_Fx = sum(i for i in self.all_log_det_Fx if i is not None)
        self.sum_pex_invFx_pex = sum(i for i in self.all_pex_invFx_pex if i is not None)
        #
        self._calculate_variance_scale()
        #
        self.neg_log_likelihood_contributions = [
            (log_det_Fx + pex_invFx_pex + num_obs * self.LOG_2_PI)/2 if num_obs else 0
            for log_det_Fx, pex_invFx_pex, num_obs in zip(
                self.all_log_det_Fx, self.all_pex_invFx_pex, self.all_num_obs,
            )
            if num_obs is not None
        ]
        #
        self.neg_log_likelihood = (
            + self.sum_num_obs*self.LOG_2_PI
            + self.sum_log_det_Fx
            + self.sum_pex_invFx_pex
        ) / 2;

    def _calculate_variance_scale(self: Self, /, ) -> None:
        """
        """
        if not self.needs.rescale_variance:
            self.var_scale = 1
            return
        if self.sum_num_obs == 0:
            self.var_scale = 1
            self.sum_pex_invFx_pex = 0
            return
        self.var_scale = self.sum_pex_invFx_pex / self.sum_num_obs
        self.sum_log_det_Fx += self.sum_num_obs * _np.log(self.var_scale)
        self.sum_pex_invFx_pex = self.sum_pex_invFx_pex / self.var_scale

    def create_output_info(self: Self, /, ) -> dict[str, Any]:
        """
        """
        return {
            "neg_log_likelihood": self.neg_log_likelihood,
            "neg_log_likelihood_contributions": self.neg_log_likelihood_contributions,
            "var_scale": self.var_scale,
            "diffuse_factor": self.diffuse_factor,
            "anticipated_shocks_impact": self.all_v_impact,
        }

    #]


def _sqrt_positive(x):
    return _np.sqrt(_np.maximum(x, 0.0, ))


def _std_from_mse(mse, ):
    return _sqrt_positive(_np.diag(mse).copy())


class Mixin:
    """
    """
    #[

    @_pages.reference(category="filtering", )
    def kalman_filter(
        self,
        #
        input_db: Databox,
        span: Iterable[Dater],
        #
        diffuse_factor: Real | None = None,
        #
        return_: Iterable[str, ...] = ("predict", "update", "smooth", ),
        return_predict: bool = True,
        return_update: bool = True,
        return_smooth: bool = True,
        rescale_variance: bool = False,
        #
        shocks_from_data: bool = False,
        stds_from_data: bool = False,
        #
        prepend_initial: bool = False,
        append_terminal: bool = False,
        #
        unpack_singleton: bool = True,
    ) -> Databox:
        r"""
················································································

==Run Kalman filter on a model using time series data==

Executes a Kalman filter on a model, compliant with `KalmanFilterableProtocol`, 
using time series observations from the input Databox. This method enables state 
estimation and uncertainty quantification in line with the model's dynamics and 
the time series data.

    kalman_output, output_info = self.kalman_filter(
        input_db, 
        span, 
        diffuse_factor=None, 
        return_=("predict", "update", "smooth"),
        return_predict=True, 
        return_update=True, 
        return_smooth=True, 
        rescale_variance=False,
        shocks_from_data=False, 
        stds_from_data=False, 
        prepend_initial=False,
        append_terminal=False, 
        unpack_singleton=True
    )


### Input arguments ###


???+ input "self"
    The model, compliant with `KalmanFilterableProtocol`, performing the 
    Kalman filtering.

???+ input "input_db"
    A Databox containing time series data to be used for filtering.

???+ input "span"
    A date span over which the filtering process is executed based on the
    measurement time series.

???+ input "diffuse_factor"
    A real number or `None`, specifying the scale factor for the diffuse
    initialization. If `None`, the default value is used.

???+ input "return_"
    An iterable of strings indicating which steps' results to return: 
    "predict", "update", "smooth".

???+ input "return_predict"
    If `True`, return prediction step results.

???+ input "return_update"
    If `True`, return update step results.

???+ input "return_smooth"
    If `True`, return smoothing step results.

???+ input "rescale_variance"
    If `True`, rescale all variances by the optimal variance scale factor
    estimated using maximum likelihood after the filtering process.

???+ input "shocks_from_data"
    If `True`, use possibly time-varying shock values from the data; these
    values are interpreted as the medians (means) of the shocks. If `False`,
    zeros are used for all shocks.

???+ input "stds_from_data"
    If `True`, use possibly time-varying standard deviation values from the
    data. If `False`, currently assigned constant values are used for the
    standard deviations of all shocks.

???+ input "prepend_initial"
    If `True`, prepend observations to the resulting time series to cover
    initial conditions based on the model's maximum lag. No measurement
    observations are used in these initial time periods (even if some are
    available in the input data).

???+ input "append_terminal"
    If `True`, append observations to the resulting time series to cover
    terminal conditions based on the model's maximum lead. No measurement
    observations are used in these terminal time periods (even if some are
    available in the input data).

???+ input "unpack_singleton"
    If `True`, unpack `output_info` into a plain dictionary for models with a
    single variant.


### Returns ###


???+ returns "kalman_output"
    An object containing the following attributes, each being a Databox:

    | Attribute                  | Description
    |----------------------------|---------------------------------------------------
    | `predict_med`              | Medians from the prediction step.
    | `predict_std`              | Standard deviations from the prediction step.
    | `predict_mse_measurement`  | Mean squared error matrices from the prediction step.
    | `update_med`               | Medians from the update step.
    | `update_std`               | Standard deviations from the update step.
    | `predict_err`              | Prediction errors.
    | `smooth_med`               | Medians from the smoothing step.
    | `smooth_std`               | Standard deviations from the smoothing step.

    Some of these attributes may be `None` if the corresponding step was not
    requested in `return_`.

???+ returns "output_info"
    A dictionary containing additional information about the filtering process,
    such as log likelihood and variance scale. For models with multiple
    variants, `output_info` is a list of such dictionaries. If
    `unpack_singleton=False`, also `output_info` is a one-element list
    containing the dictionary for singleton models, too.

················································································
        """
        work_db = input_db.shallow()
        #if not shocks_from_data:
        #    shock_names = self.get_names(kind=_quantities.ANY_SHOCK, )
        #    work_db.remove(shock_names, strict_names=False, )
        #if not stds_from_data:
        #    std_names = self.get_names(kind=_quantities.ANY_STD, )
        #    work_db.remove(std_names, strict_names=False, )

        slatable = self.get_slatable(
            shocks_from_data=shocks_from_data,
            stds_from_data=stds_from_data,
        )

        input_ds = Dataslate.from_databox_for_slatable(
            slatable, work_db, span,
            prepend_initial=prepend_initial,
            append_terminal=append_terminal,
            clip_data_to_base_span=True,
        )

        needs = _Needs(
            return_predict=return_predict and "predict" in return_,
            return_update=return_update and "update" in return_,
            return_smooth=return_smooth and "smooth" in return_,
            rescale_variance=rescale_variance,
        )

        solution_vectors = self.solution_vectors
        qid_pos_tuples = [
            (tok.qid, pos)
            for pos, tok in enumerate(solution_vectors.transition_variables, )
            if not tok.shift
        ]
        curr_qids, curr_pos = tuple(zip(*qid_pos_tuples))
        curr_qids = list(curr_qids)
        curr_pos = list(curr_pos)

        y_qids = [ t.qid for t in solution_vectors.measurement_variables ]
        u_qids = [ t.qid for t in solution_vectors.unanticipated_shocks ]
        v_qids = [ t.qid for t in solution_vectors.anticipated_shocks ]
        w_qids = [ t.qid for t in solution_vectors.measurement_shocks ]

        qid_to_logly = self.create_qid_to_logly()
        logly_within_y = tuple( i for i, qid in enumerate(y_qids, ) if qid_to_logly.get(qid, False) )

        std_u_qids = self.get_std_qids_for_shock_qids(u_qids, )
        std_w_qids = self.get_std_qids_for_shock_qids(w_qids, )

        num_variants = 1

        qid_to_name = self.create_qid_to_name()
        y_names = [ qid_to_name[qid] for qid in y_qids ]

        name_to_log_name = {
            n: _quantities.wrap_logly(n, )
            for qid, n in enumerate(input_ds.names, )
            if qid_to_logly.get(qid, False, )
        } if qid_to_logly else {}

        output_store = _OutputStore(
            input_ds=input_ds,
            num_variants=0,
            measurement_names=y_names,
            needs=needs,
            name_to_log_name=name_to_log_name,
        )

        output_info = []

        create_cache = _ft.partial(
            _Cache,
            needs=needs,
            curr_qids=curr_qids,
            curr_pos=curr_pos,
            y_qids=y_qids,
            u_qids=u_qids,
            v_qids=v_qids,
            w_qids=w_qids,
            std_u_qids=std_u_qids,
            std_w_qids=std_w_qids,
            logly_within_y=logly_within_y,
        )


        # TODO: Iterate over variants

        #
        # Variant dependent from here
        #
        input_ds_v = input_ds.get_variant(0, )
        self_v = self.get_variant(0, )

        output_store_v = _OutputStore(
            input_ds=input_ds_v,
            num_variants=1,
            measurement_names=y_names,
            needs=needs,
            name_to_log_name=name_to_log_name,
        )

        cache_v = create_cache(
            model_v=self_v,
            input_ds_v=input_ds_v,
            solution=self_v.get_solution(),
        )

        #
        # Initialize medians and stds
        #
        init_cov_u = self_v.get_cov_unanticipated_shocks()
        init_med, init_mse, cache_v.diffuse_factor = _ford_initializers.initialize(
            cache_v.solution,
            init_cov_u,
            diffuse_scale=diffuse_factor,
        )

        # tolerance = 1e-12
        # Za[_np.abs(Za)<tolerance] = 0
        # Ua[_np.abs(Ua)<tolerance] = 0
        # Pa[_np.abs(Pa)<tolerance] = 0

        a1, P1 = init_med, init_mse
        Ta = cache_v.solution.Ta
        Ra = cache_v.solution.Ra
        Pa = cache_v.solution.Pa
        Ka = cache_v.solution.Ka
        Za = cache_v.solution.Za
        H = cache_v.solution.H
        D = cache_v.solution.D
        Ua = cache_v.solution.Ua


        if needs.return_predict:
            output_store_v.predict_med.store(cache_v.u0_array, (cache_v.u_qids, ...), )
            output_store_v.predict_med.store(cache_v.v0_array, (cache_v.v_qids, ...), )
            output_store_v.predict_med.store(cache_v.w0_array, (cache_v.w_qids, ...), )

        if needs.return_update:
            output_store_v.update_med.store(cache_v.y1_array, (cache_v.y_qids, ...), )

        if needs.return_smooth:
            pass

        for t in range(cache_v.num_periods, ):

            inx_y = cache_v.all_inx_y[t]

            # Extract time-varying means and stds for shocks
            u0 = cache_v.u0_array[:, t]
            v0 = cache_v.v0_array[:, t]
            w0 = cache_v.w0_array[:, t]
            cov_u = _np.diag(cache_v.std_u_array[:, t]**2, )
            cov_w = _np.diag(cache_v.std_w_array[:, t]**2, )
            v_impact = cache_v.all_v_impact[t] if cache_v.all_v_impact[t] is not None else 0

            #
            # MSE prediction step
            #
            P0 = Ta @ P1 @ Ta.T + Pa @ cov_u @ Pa.T
            P0 = (P0 + P0.T) / 2
            F = Za @ P0 @ Za.T + H @ cov_w @ H.T
            F = (F + F.T) / 2
            Fx = F[inx_y, :][:, inx_y]
            if needs.return_predict:
                output_store_v.predict_std.store(P0, (cache_v.curr_qids, t), rhs_indexes=cache_v.curr_pos, transform=Ua, )
                output_store_v.predict_mse_measurement[0] += (Fx, )

            #
            # Mean prediction step
            #
            a0 = Ta @ a1 + Ka + Pa @ u0 + v_impact
            y0 = Za @ a0 + D + H @ w0
            y0[~inx_y] = _np.nan

            if needs.return_predict:
                output_store_v.predict_med.store(a0, (cache_v.curr_qids, t), rhs_indexes=cache_v.curr_pos, transform=Ua, )
                output_store_v.predict_med.store(y0, (cache_v.y_qids, t), )
                output_store_v.predict_med.store(u0, (cache_v.u_qids, t), )
                output_store_v.predict_med.store(v0, (cache_v.v_qids, t), )
                output_store_v.predict_med.store(w0, (cache_v.w_qids, t), )

            #
            # MSE updating step
            #
            if Fx.size:
                # sing_values = _np.linalg.svd(Fx, compute_uv=False, )
                # relative_cond = sing_values[-1] / sing_values[0]
                # print(t, Fx.size, relative_cond, sing_values[0], )
                # TODO: Check if Fx singularity if requested by user
                invFx = _np.linalg.inv(Fx)
            else:
                invFx = _np.zeros((0, 0), dtype=_np.float64, )
            #
            Zax = Za[inx_y, :]
            # ZaxT_invFx = right_div(Zax.T, Fx, )
            #
            ZaxT_invFx = Zax.T @ invFx
            G = P0 @ ZaxT_invFx
            P1 = P0 - G @ Zax @ P0
            P1 = (P1 + P1.T) / 2

            if needs.return_update:
                output_store_v.update_std.store(P1, (cache_v.curr_qids, t), rhs_indexes=cache_v.curr_pos, transform=Ua, )

            #
            # Mean updating step
            #
            cache_v.all_num_obs[t] = _np.count_nonzero(inx_y, )
            y1 = cache_v.y1_array[:, t]
            pe = y1 - y0
            pex = pe[inx_y]
            a1 = a0 + G @ pex

            # invFx_pex = left_div(Fx, pex, ) # inv(Fx) * pex
            invFx_pex = invFx @ pex # inv(Fx) * pex

            if needs.return_update:
                Hx_cov_w = H[inx_y, :] @ cov_w
                w1 = w0 + Hx_cov_w.T @ invFx_pex
                Pa_cov_u = Pa @ cov_u
                r = ZaxT_invFx @ pex # Z' * inv(F) * pe
                u1 = u0 + Pa_cov_u.T @ r
                v1 = v0
                output_store_v.update_med.store(a1, (cache_v.curr_qids, t), rhs_indexes=cache_v.curr_pos, transform=Ua, )
                output_store_v.update_med.store(y1, (cache_v.y_qids, t), )
                output_store_v.update_med.store(u1, (cache_v.u_qids, t), )
                output_store_v.update_med.store(v1, (cache_v.v_qids, t), )
                output_store_v.update_med.store(w1, (cache_v.w_qids, t), )
                output_store_v.predict_err.store(pe, (..., t), )

            if needs.return_smooth:
                cache_v.all_a0[t] = a0
                cache_v.all_P0[t] = P0
                cache_v.all_a1[t] = a1
                cache_v.all_P1[t] = P1
                cache_v.all_invFx_pex[t] = invFx_pex
                cache_v.all_ZaxT_invFx[t] = Zax.T @ invFx
                cache_v.all_ZaxT_invFx_Zax[t] = ZaxT_invFx @ Zax
                cache_v.all_L[t] = Ta - (Ta @ G) @ Zax
                cache_v.all_G[t] = G
                cache_v.all_cov_u[t] = cov_u
                cache_v.all_cov_w[t] = cov_w

            cache_v.all_pex[t] = pex
            cache_v.all_invFx[t] = invFx
            cache_v.all_pex_invFx_pex[t] = pex @ invFx_pex
            cache_v.all_log_det_Fx[t] = _np.log(_np.linalg.det(Fx))


        if needs.return_smooth:
            _smooth_backward(cache_v, output_store_v, )

        cache_v.calculate_likelihood()
        output_store_v.rescale_stds(cache_v.var_scale, )
        output_info_v = cache_v.create_output_info()

        output_store.extend(output_store_v, )
        output_info.append(output_info_v, )

        output_info = _has_variants.unpack_singleton(
            output_info, self.is_singleton,
            unpack_singleton=True,
        )

        return output_store.create_output_data(), output_info

    #]


def _smooth_backward(
    cache_v: _Cache,
    output_store_v: _OutputStore,
) -> None:
    """
    """
    #[

    Ta = cache_v.solution.Ta
    Ua = cache_v.solution.Ua
    Pa = cache_v.solution.Pa
    H = cache_v.solution.H

    N = None
    r = None

    for t in reversed(range(cache_v.num_periods, )):

        inx_y = cache_v.all_inx_y[t]
        u0 = cache_v.u0_array[:, t]
        v0 = cache_v.v0_array[:, t]
        w0 = cache_v.w0_array[:, t]
        cov_u = cache_v.all_cov_u[t]
        cov_w = cache_v.all_cov_w[t]

        a0 = cache_v.all_a0[t]
        L = cache_v.all_L[t]
        G = cache_v.all_G[t]
        P0 = cache_v.all_P0[t]
        ZaxT_invFx = cache_v.all_ZaxT_invFx[t]
        ZaxT_invFx_Zax = cache_v.all_ZaxT_invFx_Zax[t]
        invFx = cache_v.all_invFx[t]
        pex = cache_v.all_pex[t]

        N = ZaxT_invFx_Zax + ((L.T @ N @ L) if N is not None else 0)
        P2 = P0 - P0 @ N @ P0
        output_store_v.smooth_std.store(P2, (cache_v.curr_qids, t), rhs_indexes=cache_v.curr_pos, transform=Ua, )

        H_cov_w = H[inx_y, :] @ cov_w
        w2 = w0 + H_cov_w.T @ (invFx @ pex - (((Ta @ G).T @ r) if r is not None else 0))
        output_store_v.smooth_med.store(w2, (cache_v.w_qids, t), )

        r = ZaxT_invFx @ pex + ((L.T @ r) if r is not None else 0)
        a2 = a0 + P0 @ r
        output_store_v.smooth_med.store(a2, (cache_v.curr_qids, t), rhs_indexes=cache_v.curr_pos, transform=Ua, )

        y2 = cache_v.y1_array[:, t]
        output_store_v.smooth_med.store(y2, (cache_v.y_qids, t), )

        Pa_cov_u = Pa @ cov_u
        u2 = u0 + Pa_cov_u.T @ r
        output_store_v.smooth_med.store(u2, (cache_v.u_qids, t), )

        v2 = v0
        output_store_v.smooth_med.store(v2, (cache_v.v_qids, t), )

    #]




"""
"""


#[
from __future__ import annotations

from typing import (Self, Any, )
import prettytable as _pt
import itertools as _it

from .. import pages as _pages
#]


class PrettyMixin:
    @property
    def pretty(self, /, ) -> _pt.PrettyTable:
        """
        """
        return self.get_pretty()

    @property
    def pretty_full(self, /, ) -> str:
        """
        """
        return self.get_pretty(full=True, )

    def get_pretty(
        self,
        /,
    ) -> _pt.PrettyTable:
        """
        """
        table = _pt.PrettyTable()
        table.field_names = ("NAME", "PERIOD(S)", "REGISTER", "TRANSFORM", )
        table.align = "r"
        table.align["NAME"] = "l"
        for r in ("exogenized", "endogenized", ):
            if getattr(self, f"_{r}_register"):
                _add_register_to_table(
                    table,
                    self.base_span,
                    getattr(self, f"_{r}_register"),
                    r,
                )
        return table

    @_pages.reference(
        category="information",
    )
    def print_table(self, *args, **kwargs, ) -> None:
        r"""
................................................................................

==Print the `SimulationPlan` as a table==

    self.print_table()


### Input arguments ###

???+ input "self"
    `SimulationPlan` to be printed on the screen, with one row showing
    exogenized or endogenized data points grouped by the name and dates.


### Returns ###

Returns no value; the table is printed on the screen.

................................................................................
        """
        print(self.get_pretty(*args, **kwargs, ), )

    def pretty_print(self, *args, **kwargs, ) -> None:
        """
        """
        print(self.get_pretty(*args, **kwargs, ), )

    def get_pretty_string(self, *args, **kwargs, ) -> str:
        """
        """
        return self.get_pretty(*args, **kwargs, ).get_string()


def _add_register_to_table(
    table,
    base_span,
    register,
    action,
) -> None:
    """
    """
    #[
    def create_row(name, date, action, status, ):
        return (
            name,
            str(date),
            action,
            status.symbol if hasattr(status, "symbol") else _PRETTY_SYMBOL.get(status, ""),
        )
    def create_representative(rows, ):
        if len(rows) == 1:
            return rows[0]
        else:
            return (rows[0][0], rows[0][1] + ">>" + rows[-1][1], rows[0][2], rows[0][3], )
    #
    all_rows = (
        create_row(k, date, action, status, )
        for k, v in register.items()
        for status, date in zip(v, base_span)
        if status is not None and status is not False
    )
    all_rows = sorted(all_rows, key=lambda row: (row[0], row[1], ), )
    row_groups = _it.groupby(all_rows, key=lambda row: (row[0], row[2], row[3], ), )
    for _, g in row_groups:
        representative = create_representative(tuple(g), )
        table.add_row(representative, )
    #]



_PRETTY_SYMBOL = {
    None: "",
    True: "*",
    False: "",
}





"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, Protocol, Callable, )
import numpy as _np
import functools as _ft
import wlogging as _wl

from .. import frames as _frames
from .. import dates as _dates
from ..dataslates import main as _dataslates
from ..plans import main as _plans
from ..fords import solutions as _solutions
from ..fords import descriptors as _descriptors
#]


class _FordSimulatableProtocol:
    """
    """
    #[

    num_variants: int

    solution_vectors: _descriptors.SolutionVectors

    def get_solution(self, /, ) -> _solutions.Solution: ...

    #]


def simulate(
    model_v: _FordSimulatableProtocol,
    dataslate_v: _dataslates.Dataslate,
    plan: _plans.SimulationPlan | None,
    vid: int,
    logger: _wl.Logger,
    /,
    *,
    deviation: bool = False,
) -> dict[str, Any]:
    """
    """
    #[
    orig_base_periods = dataslate_v.base_periods
    needs_split = plan is not None and plan.needs_split
    create_frames = (
        _frames.split_into_frames if needs_split
        else _frames.setup_single_frame
    )
    frames = create_frames(
        model_v, dataslate_v, plan,
        get_simulation_end=lambda start, end: end,
    )
    #
    for frame in frames:
        dataslate_v.base_periods \
            = _dates.periods_from_to(frame.start, frame.simulation_end, )
        simulate_flat(
            model_v, dataslate_v, vid,
            plan=plan,
            deviation=deviation,
        )
        dataslate_v.base_periods = orig_base_periods
    #
    info = {
        "method": "first_order",
        "frames": frames,
    }
    #
    return info
    #]


def simulate_flat(
    model_v: _FordSimulatableProtocol,
    dataslate_v: _dataslates.Dataslate,
    vid: int,
    /,
    *,
    plan: _plans.PlanSimulate | None = None,
    deviation: bool = False,
) -> None:
    """
    """
    #[
    if model_v.num_variants != 1:
        raise ValueError("Simulator requires a singleton simulatable object")
    #
    solution = model_v.get_solution()
    solution_vectors = model_v.solution_vectors
    columns_to_run = dataslate_v.base_columns
    logly_indexes = dataslate_v.logly_indexes
    working_data = dataslate_v.get_data_variant(0, )
    #
    first_column = columns_to_run[0]
    last_column = columns_to_run[-1]
    column_range = tuple(range(first_column, last_column+1))
    #
    vec = solution_vectors
    #
    T = solution.T
    P = solution.P
    R = solution.R
    K = solution.K if not deviation else 0
    #
    Z = solution.Z
    H = solution.H
    D = solution.D if not deviation else 0
    #
    if logly_indexes:
        working_data[logly_indexes, ...] = _np.log(working_data[logly_indexes, ...])
    #
    curr_state = _dataslates.retrieve_vector_from_data_array(
        working_data, vec.transition_variables, first_column-1,
    )
    #
    missing_initials = _np.isnan(curr_state, )
    unnecessary_initials = ~_np.array(vec.are_initial_conditions, )
    missing_unnecessary_initials = missing_initials & unnecessary_initials
    curr_state[missing_unnecessary_initials] = 0
    #
    no_shift_state_to_slab_lhs = [
        t.qid
        for t in vec.transition_variables
        if t.shift == 0
    ]
    #
    no_shift_state_to_slab_rhs = [
        j 
        for j, t in enumerate(vec.transition_variables)
        if t.shift == 0
    ]
    #
    measurement_variables_index = [t.qid for t in vec.measurement_variables]
    #
    u = _extract_shock_values(working_data, vec.unanticipated_shocks, )
    w = _extract_shock_values(working_data, vec.measurement_shocks, )
    #
    v_impact \
        = simulate_square_anticipated_shocks(model_v, dataslate_v, )
    #
    for t in column_range:
        curr_state = T @ curr_state + P @ u[:, t] + K
        if v_impact[t] is not None:
            curr_state += v_impact[t]
        #
        working_data[no_shift_state_to_slab_lhs, t] = curr_state[no_shift_state_to_slab_rhs]
        #
        y = Z @ curr_state + H @ w[:, t] + D
        working_data[measurement_variables_index, t] = y
        #
    if logly_indexes:
        working_data[logly_indexes, ...] = _np.exp(working_data[logly_indexes, ...])
    #
    info = {
        "method": "first_order",
        "anticipatory_shocks_impact": v_impact,
    }
    #
    return info
    #]


def _extract_shock_values(working_data, shock_vector, ) -> _np.ndarray:
    """
    """
    #[
    shock_qids = [t.qid for t in shock_vector]
    shock_values = working_data[shock_qids, :]
    return shock_values
    #]


def _simulate_anticipated_shocks(
    model_v: _FordSimulatableProtocol,
    dataslate_v: _dataslates.Dataslate,
    expand_solution: Callable,
) -> list[_np.ndarray | None]:
    """
    """
    #[
    if model_v.num_variants != 1:
        raise ValueError("Simulator requires a singleton simulatable object")
    #
    solution = model_v.get_solution()
    solution_vectors = model_v.solution_vectors
    base_columns = dataslate_v.base_columns
    first_column = base_columns[0]
    last_column = base_columns[-1]
    column_range = tuple(range(first_column, last_column+1))
    working_data = dataslate_v.get_data_variant(0, )
    anticipated_shocks = _extract_shock_values(working_data, solution_vectors.anticipated_shocks, )
    #
    impact = [None] * dataslate_v.num_periods
    column_incidence = _np.any(anticipated_shocks[:, column_range], axis=0, ).tolist()
    if any(column_incidence, ):
        #
        # Last shock is at t+forward
        forward = len(column_incidence) - column_incidence[::-1].index(True, ) - 1
        Rx = expand_solution(solution, forward, )
        #
        # Last column within dataslate_v
        last_ant_column = (
            first_column + forward
            if forward is not None
            else first_column - 1
        )
        #
        for t in column_range:
            impact[t] = sum(
                Rx[k] @ anticipated_shocks[:, s]
                for k, s in enumerate(range(t, last_ant_column+1))
            )
    return impact
    #]


simulate_square_anticipated_shocks = _ft.partial(
    _simulate_anticipated_shocks,
    expand_solution=_solutions.Solution.expand_square_solution,
)


simulate_triangular_anticipated_shocks = _ft.partial(
    _simulate_anticipated_shocks,
    expand_solution=_solutions.Solution.expand_triangular_solution,
)


def _get_frames(
    model_v: _FordSimulatableProtocol,
    dataslate_v: _dataslates.Dataslate,
    plan: _plans.PlanSimulate | None = None,
) -> None:
    """
    """





"""
Implement SlatableProtocol
"""


#[
from __future__ import annotations

from .. import quantities as _quantities
from ..series.main import (Series, )
#]


class _Slatable:
    """
    """
    #[

    def __init__(
        self,
        simultaneous,
        *,
        shocks_from_data: bool = False,
        stds_from_data: bool = False,
        **kwargs,
    ) -> None:
        """
        """
        #
        # Min and max shifts
        self.max_lag = simultaneous.max_lag
        self.max_lead = simultaneous.max_lead
        #
        # Databox names
        qid_to_name = simultaneous.create_qid_to_name()
        self.databox_names = tuple(
            qid_to_name[qid]
            for qid in sorted(qid_to_name)
        )
        #
        # Databox validation - all variables must be time series
        variable_names = simultaneous.get_names(kind=_quantities.ANY_VARIABLE, )
        validator = (
            lambda x: isinstance(x, Series),
            "Data for this variable is not a time series",
        )
        self.databox_validators = {
            name: validator
            for name in variable_names
        }
        #
        # Fallbacks and overwrites
        self.fallbacks = {}
        self.overwrites = simultaneous.get_parameters(unpack_singleton=False, )
        #
        num_variants = simultaneous.num_variants
        shock_names = simultaneous.get_names(kind=_quantities.ANY_SHOCK, )
        shock_meds = { name: [float(0), ]*num_variants for name in shock_names }
        shock_stds = simultaneous.get_stds(unpack_singleton=False, )
        #
        if shocks_from_data:
            self.fallbacks.update(shock_meds, )
        else:
            self.overwrites.update(shock_meds, )
        #
        if stds_from_data:
            self.fallbacks.update(shock_stds, )
        else:
            self.overwrites.update(shock_stds, )
        #
        # Qid to logly
        self.qid_to_logly = simultaneous.create_qid_to_logly()
        #
        # Output names
        kind = _quantities.ANY_VARIABLE | _quantities.ANY_SHOCK
        self.output_names = simultaneous.get_names(kind=kind, )


class Inlay:
    """
    """
    #[

    def get_slatable(self, **kwargs, ) -> _Slatable:
        """
        """
        return _Slatable(self, **kwargs, )





"""
Implement SimulationPlannableProtocol and SteadyPlannableProtocol
"""


#[
from __future__ import annotations

from .. import quantities as _quantities
#]


_SIMULATE_CAN_BE_EXOGENIZED = _quantities.QuantityKind.ENDOGENOUS_VARIABLE
_SIMULATE_CAN_BE_ENDOGENIZED = _quantities.QuantityKind.EXOGENOUS_VARIABLE | _quantities.QuantityKind.ANY_SHOCK
_SIMULATE_CAN_BE_EXOGENIZED_ANTICIPATED = _quantities.QuantityKind.ENDOGENOUS_VARIABLE
_SIMULATE_CAN_BE_ENDOGENIZED_ANTICIPATED = _quantities.QuantityKind.ANTICIPATED_SHOCK
_SIMULATE_CAN_BE_EXOGENIZED_UNANTICIPATED = _quantities.QuantityKind.ENDOGENOUS_VARIABLE
_SIMULATE_CAN_BE_ENDOGENIZED_UNANTICIPATED = _quantities.QuantityKind.UNANTICIPATED_SHOCK


class _SimulationPlannable:
    """
    """
    #[

    def __init__(
        self,
        simultaneous,
        **kwargs,
    ) -> None:
        """
        """
        self.can_be_exogenized_anticipated \
            = tuple(_quantities.generate_quantity_names_by_kind(
                simultaneous._invariant.quantities,
                _SIMULATE_CAN_BE_EXOGENIZED_ANTICIPATED,
            ))
        self.can_be_exogenized_unanticipated \
            = tuple(_quantities.generate_quantity_names_by_kind(
                simultaneous._invariant.quantities,
                _SIMULATE_CAN_BE_EXOGENIZED_UNANTICIPATED,
            ))
        self.can_be_endogenized_anticipated \
            = tuple(_quantities.generate_quantity_names_by_kind(
                simultaneous._invariant.quantities,
                _SIMULATE_CAN_BE_ENDOGENIZED_ANTICIPATED,
            ))
        self.can_be_endogenized_unanticipated \
            = tuple(_quantities.generate_quantity_names_by_kind(
                simultaneous._invariant.quantities,
                _SIMULATE_CAN_BE_ENDOGENIZED_UNANTICIPATED,
            ))


class Inlay:
    """
    """
    #[

    def get_simulation_plannable(self, **kwargs, ) -> _SimulationPlannable:
        """
        """
        return _SimulationPlannable(self, **kwargs, )

    #]




"""
Plans for dynamic simulations and steady state calculations
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Self, Any, Protocol, NoReturn, )
from types import (EllipsisType, )
import warnings as _wa
import functools as _ft
import numpy as _np
import copy as _copy

from ..conveniences import copies as _copies
from ..dates import (Period, )
from .. import wrongdoings as _wrongdoings
from .. import pages as _pages
from . import _pretty as _pretty
from . import _indexes as _indexes
from . import transforms as _transforms
#]


CHOOSE_TRANSFORM_CLASS = _transforms.CHOOSE_TRANSFORM_CLASS


__all__ = (
    "SimulationPlan", "PlanSimulate", "Plan",
    "PlanSteady", "SteadyPlan",
    "CHOOSE_TRANSFORM_CLASS",
)


class SimulationPlannableProtocol(Protocol, ):
    """
    """
    #[

    simulate_can_be_exogenized: Iterable[str] | None
    simulate_can_be_endogenized: Iterable[str] | None
    simulate_can_be_exogenized_anticipated: Iterable[str] | None
    simulate_can_be_exogenized_anticipated: Iterable[str] | None
    simulate_can_be_endogenized_unanticipated: Iterable[str] | None
    simulate_can_be_endogenized_unanticipated: Iterable[str] | None

    #]


class SteadyPlannableProtocol(Protocol, ):
    """
    """
    #[

    steady_can_be_exogenized: Iterable[str] | None
    steady_can_be_endogenized: Iterable[str] | None
    steady_can_be_fixed_level: Iterable[str] | None
    steady_can_be_fixed_change: Iterable[str] | None

    #]


@_pages.reference(
    path=("structural_models", "simulation_plans.md", ),
    categories={
        "constructor": "Creating new simulation plans",
        "definition_simultaneous": "Defining exogenized and endogenized data points in [`Simultaneous` simulations](simultaneous.md#simulate)",
        "definition_sequential": "Defining exogenized and endogenized data points in [`Sequential` simulations](sequential.md#simulate)",
        "information": "Getting information about simulation plans",
        "information_simultaneous": "Getting information about simulation plans for [`Simultaneous` models](simultaneous.md)",
        "information_sequential": "Getting information about simulation plans for [`Sequential` models](sequential.md)",
    },
)
class SimulationPlan(
    _pretty.PrettyMixin,
    _indexes.ItemMixin,
    _copies.CopyMixin,
):
    """
················································································

Simulation plans
=================

`SimulationPlan` objects are used to set up conditioning assumptions
for simulating [`Simultaneous`](simultaneous_modelsd) or
[`Sequential`](sequential_models) models. The simulation plans specify

* what variables to exogenize in what periods
* what shocks to endogenized in what periods (`Simultaneous` models only)

The plans only contain meta information, not the actual data points for the
exogenized variables. The actual data points are included in the input databox.

················································································
    """
    #[

    _registers = (
        "exogenized",
        "endogenized",
        "exogenized_anticipated",
        "exogenized_unanticipated",
        "endogenized_unanticipated",
        "endogenized_anticipated",
    )

    __slots__ = (
        *tuple(f"can_be_{r}" for r in _registers),
        *tuple(f"_{r}_register" for r in _registers),
        *tuple(f"default_{r}" for r in _registers),
        "base_span",
    )

    @_pages.reference(category="constructor", call_name="SimulationPlan", )
    def __init__(
        self,
        model,
        span: Iterable[Period] | None,
    ) -> None:
        """
················································································

==Create new simulation plan object==

```
self = SimulationPlan(model, time_span, )
```

Create a new simulation plan object for a
[`Simultaneous`](sequential_models) or
[`Sequential`](sequential_models) model.

### Input arguments ###

???+ input "model"

    A [`Simultaneous`](sequential_models) or
    [`Sequential`](sequential_models) model that will be simulated.

???+ input "time_span"

    A date range on which the `model` will be simulated.


### Returns ###

???+ returns "self"

    A new empty simulation plan object.

················································································
        """
        self.base_span = tuple(span)
        self._default_exogenized = None
        self._default_endogenized = None
        plannable = model.get_simulation_plannable()
        for r in self._registers:
            can_be_name = f"can_be_{r}"
            register = {
                n: [None] * self.num_periods
                for n in getattr(plannable, can_be_name, )
            } if hasattr(plannable, can_be_name, ) else {}
            setattr(self, can_be_name, tuple(register.keys()))
            setattr(self, f"_{r}_register", register)

    def check_consistency(
        self,
        plannable: SimulationPlannableProtocol,
        span: Iterable[Period] | None,
        /,
    ) -> None:
        """
        """
        benchmark = type(self)(plannable, span, )
        if self.base_span != benchmark.base_span:
            raise _wrongdoings.IrisPieError(f"Plan span must be the same as the simulation span")
        for r in self._registers:
            if getattr(self, f"can_be_{r}") != getattr(benchmark, f"can_be_{r}"):
                raise _wrongdoings.IrisPieError(f"Plan must be created using the simulated model")

    @property
    @_pages.reference(category="property", )
    def start_date(self, /, ) -> Period:
        """==Start date of the simulation span=="""
        return self.base_span[0]

    @property
    @_pages.reference(category="property", )
    def end_date(self, /, ) -> Period:
        """==End date of the simulation span=="""
        return self.base_span[-1]

    @property
    @_pages.reference(category="property", )
    def num_periods(self, /, ) -> int:
        """==Number of periods in the simulation span=="""
        return len(self.base_span) if self.base_span is not None else 1

    @property
    @_pages.reference(category="property", )
    def frequency(self, /, ) -> str:
        """==Date frequency of the simulation span=="""
        return self.start_date.frequency

    @_pages.reference(category="definition_sequential", )
    def exogenize(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        /,
        *,
        transform: str | None = None,
        # when_data: bool | None = None,
        **kwargs,
    ) -> None:
        r"""
................................................................................

==Exogenize certain LHS quantities at certain dates==

Exogenize certain LHS quantities at specified dates, setting them as
predetermined values within the simulation of
a [`Sequential` model](sequential.md). This method is used to control how
the model behaves during simulations by fixing certain variables to known
values.

    self.exogenize(
        dates,
        names,
        *,
        transform=None,
        when_data=False,
    )

### Input arguments ###

???+ input "self"
    The simulation plan in which data points will be exogenized.

???+ input "dates"
    A list of dates or `...` to apply to all dates at which the quantities 
    will be exogenized.

???+ input "names"
    A list of names or a single name, or `...` to apply to all names that 
    specifies which quantities to set as predetermined at the specified dates.

???+ input "transform"
    Specifies the transformation to apply to the exogenized quantities. If not
    specified, no transformation is applied. Available transformations include:

    * `None`: Exogenize the LHS variables as they are with no
    transformation.

    * `"log"`: Exogenize the natural logarithm of the LHS variables. Input
    time series needs to be prefixed with `log_`.

    * `"diff"`: Exogenize the first difference of the LHS variables. Input
    time series needs to be prefixed with `diff_`.

    * `"diff_log"`: Exogenize the first difference of the natural logarithm
    of the LHS variables. Input time series needs to be prefixed with
    `diff_log_`.

    * `"roc"`: The gross rate of change of the LHS variables from one
    period to the next. Input time series needs to be prefixed with `roc_`.

    * `"pct"`: The percentage change of the LHS variables from one period
    to the next. Input time series needs to be prefixed with `pct_`.

???+ input "when_data"
    Specifies whether the exogenization should only occur if a valid 
    value exists in the input data.

### Returns ###

This method modifies the object in place and does not return a value.

................................................................................
        """
        transform = _transforms.resolve_transform(transform, **kwargs, )
        self._register(self._exogenized_register, dates, names, transform, )

    @_pages.reference(category="definition_simultaneous", )
    def exogenize_anticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        *,
        status: bool | int = True,
    ) -> None:
        """
················································································

==Exogenize certain quantities at certain dates==

```
self.exogenize_anticipated(
    dates,
    names,
)
```

### Input arguments ###


???+ input "dates"

    Dates at which the `names` will be exogenized; use `...` for all simulation dates.

???+ input "names"

    Names of quantities to exogenize at the `dates`; use `...` for all exogenizable quantities.

················································································
        """
        self._register(
            self._exogenized_anticipated_register,
            dates,
            names,
            status,
        )

    @_pages.reference(category="definition_simultaneous", )
    def exogenize_unanticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        *,
        status: bool | int = True,
    ) -> None:
        r"""
················································································

==Exogenize certain quantities at certain dates as unanticipated==

```
self.exogenize_unanticipated(
    dates, names,
    /,
    transform=None,
    when_data=False,
)
```

### Input arguments ###


???+ input "dates"

    Dates at which the `names` will be exogenized; use `...` for all simulation dates.

???+ input "names"

    Names of quantities to exogenize at the `dates`; use `...` for all exogenizable quantities.


### Input arguments available only for `Sequential` models ###

???+ input "transform"

    Transformation (specified as a string) to be applied to the exogenized
    quantities; if `None`, no tranformation is applied.

???+ input "when_data"

    If `True`, the data point will be exogenized only if a proper value
    exists in the input data.

················································································
        """
        self._register(
            self._exogenized_unanticipated_register,
            dates,
            names,
            status,
        )

    @_pages.reference(category="information_sequential", )
    def get_exogenized_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names exogenized in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._exogenized_register,
            *args, **kwargs,
        )

    @_pages.reference(category="information_simultaneous", )
    def get_exogenized_unanticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names exogenized as unanticipated in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._exogenized_unanticipated_register,
            *args, **kwargs,
        )

    @_pages.reference(category="information_simultaneous", )
    def get_exogenized_anticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names exogenized as anticipated in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._exogenized_anticipated_register,
            *args, **kwargs,
        )

    # @_pages.reference(category="definition_sequential", )
    def endogenize(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        /,
    ) -> None:
        r"""
        """
        self._register(self._endogenized_register, dates, names, True, )

    @_pages.reference(category="definition_simultaneous", )
    def endogenize_anticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        *,
        status: bool | int = True,
    ) -> None:
        """
················································································

==Endogenize certain quantities at certain dates==

················································································
        """
        self._register(
            self._endogenized_anticipated_register,
            dates,
            names,
            status,
        )

    @_pages.reference(category="definition_simultaneous", )
    def endogenize_unanticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        *,
        status: bool | int = True,
    ) -> None:
        """
················································································

==Endogenize certain quantities at certain dates==

················································································
        """
        self._register(
            self._endogenized_unanticipated_register,
            dates,
            names,
            status,
        )

    @_pages.reference(category="information_simultaneous", )
    def get_endogenized_unanticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names endogenized as unanticipated in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._endogenized_unanticipated_register,
            *args,
            **kwargs,
        )

    @_pages.reference(category="information_simultaneous", )
    def get_endogenized_anticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names endogenized as anticipated in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._endogenized_unanticipated_register,
            *args,
            **kwargs,
        )

    @property
    def needs_split(self, /, ) -> bool:
        """
        """
        def _is_active_statue(value: Any, /, ) -> bool:
            return (value is not None) and (value is not False)
        return any(
            any(
                any(_is_active_statue(i) for i in v)
                for v in getattr(self, "_" + a + "_register").values()
            )
            for a in ("exogenized_unanticipated", "endogenized_unanticipated", )
        )

    def swap(
        self,
        dates: Iterable[Period] | EllipsisType,
        pairs: Iterable[tuple[str, str]] | tuple[str, str],
        *args, **kwargs,
    ) -> None:
        """
        """
        pairs = tuple(pairs)
        if not pairs:
            return
        if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
            pairs = (pairs, )
        for pair in pairs:
            self.exogenize(dates, pair[0], *args, **kwargs, )
            self.endogenize(dates, pair[1], *args, **kwargs, )

    def swap_anticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        pairs: Iterable[tuple[str, str]] | tuple[str, str],
        *args, **kwargs,
    ) -> None:
        """
        """
        pairs = tuple(pairs)
        if not pairs:
            return
        if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
            pairs = (pairs, )
        for pair in pairs:
            self.exogenize_anticipated(dates, pair[0], *args, **kwargs, )
            self.endogenize_anticipated(dates, pair[1], *args, **kwargs, )

    def swap_unanticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        pairs: Iterable[tuple[str, str]] | tuple[str, str],
        *args, **kwargs,
    ) -> None:
        """
        """
        pairs = tuple(pairs)
        if not pairs:
            return
        if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
            pairs = (pairs, )
        for pair in pairs:
            self.exogenize_unanticipated(dates, pair[0], *args, **kwargs, )
            self.endogenize_unanticipated(dates, pair[1], *args, **kwargs, )

    def tabulate_registered_points(
        self: Self,
        register_name: str,
        names: str | Iterable[str] | EllipsisType,
        periods: Iterable[Period] | EllipsisType,
    ) -> _np.ndarray | None:
        """
        """
        per_indexes = self._get_period_indexes(periods, )
        register = getattr(self, f"_{register_name}_register", )
        names = _resolve_and_check_names(register, names, )
        num_pers = len(per_indexes)
        num_names = len(names)
        return _np.array(tuple(
            tuple( bool(register[n][t]) for t in per_indexes )
            for n in names
        )) if names and per_indexes else None

    def get_exogenized_point(
        self,
        name: str,
        date: Period,
        /,
    ) -> _transforms.PlanTransformProtocol | None:
        """
        """
        column = next((
            column
            for column, t in enumerate(self.base_span)
            if date == t
        ))
        point = self._exogenized_register[name][column]
        return point if point is not None else self._default_exogenized

    def get_endogenized_point(
        self,
        name: str,
        column: int,
        /,
    ) -> _transforms.PlanTransformProtocol | None:
        """
        """
        point = self._endogenized_register[name][column]
        return point if point is not None else self._default_endogenized

    def get_databox_names(self, /, ) -> tuple[str]:
        """
        """
        databox_names = set()
        for k, v in self._exogenized_register.items():
            databox_names.update(
                t.resolve_databox_name(k, )
                for t in v if t is not None
            )
        return tuple(n for n in databox_names if n is not None)

    def _get_names_registered_in_period(
        self,
        register: dict[str, Any],
        date: Period,
    ) -> tuple[str, ...]:
        """
        """
        column_index = self.base_span.index(date)
        return tuple(
            name
            for name, status in register.items()
            if bool(status[column_index])
        )

    def __str__(self, /, ) -> str:
        """
        """
        return self.get_pretty_string()

    def _register(
        self,
        register: dict,
        periods: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        new_status: Any,
    ) -> None:
        """
        """
        names = _resolve_and_check_names(register, names, )
        per_indices = self._get_period_indexes(periods, )
        for n in names:
            for t in per_indices:
                register[n][t] = new_status

    def _get_period_indexes(
        self,
        dates: Iterable[Period] | EllipsisType,
        /,
    ) -> tuple[int, ...]:
        """
        """
        if dates is Ellipsis:
            return tuple(range(len(self.base_span)))
        dates = dates.resolve(self, ) if hasattr(dates, "resolve") else dates
        catch_invalid_dates(dates, self.base_span, )
        return tuple(d - self.start_date for d in dates)

    #]


@_pages.reference(
    path=("structural_models", "steady_plans.md", ),
    categories={
        "constructor": "Creating new steady plans",
        "property": None,
        "definition": "Defining Exogenized, endogenized and fixed quantities",
    },
)
class SteadyPlan:
    """
················································································

Steady-state plans
===================

`SteadyPlan` objects define assumptions about the steady state values of
certain model quantities (variables and parameters), either for reverse
engineered calculations or for narrowing down the underdetermined
steady-state systems.

················································································
    """
    #[

    __slots__ = (
        "can_be_exogenized",
        "can_be_endogenized",
        "can_be_fixed_level",
        "can_be_fixed_change",
        "_exogenized_register",
        "_endogenized_register",
        "_fixed_level_register",
        "_fixed_change_register",
    )

    @_pages.reference(category="constructor", call_name="SteadyPlan", )
    def __init__(
        self,
        model,
        /,
    ) -> None:
        """
················································································

==Create new steady plan object==

················································································
        """
        plannable = model.get_steady_plannable()
        for r in ("exogenized", "endogenized", "fixed_level", "fixed_change", ):
            can_be_name = f"can_be_{r}"
            register = {
                n: False
                for n in getattr(plannable, can_be_name, )
            } if hasattr(plannable, can_be_name, ) else {}
            setattr(self, can_be_name, tuple(register.keys()))
            setattr(self, f"_{r}_register", register)

    #]

#         self.exogenized = {
#             n: [None] * self.num_periods
#             for n in plannable.can_be_exogenized
#         } if plannable.can_be_exogenized else {}
#         #
#         self.endogenized = {
#             n: [None] * self.num_periods
#             for n in plannable.can_be_endogenized
#         } if plannable.can_be_endogenized else {}
#         #
#         self.fixed_level = {
#             n: [None] * self.num_periods
#             for n in plannable.can_be_fixed_level
#         } if plannable.can_be_fixed_level else {}
#         #
#         self.fixed_change = {
#             n: [None] * self.num_periods
#             for n in plannable.can_be_fixed_change
#         } if plannable.can_be_fixed_change else {}

    @_pages.reference(category="definition", )
    def exogenize(
        self,
        names: Iterable[str] | str,
        /,
    ) -> None:
        """
················································································

==Exogenize steady levels of variables==

················································································
        """
        pass

    @_pages.reference(category="definition", )
    def endogenize(
        self,
        names: Iterable[str] | str,
        /,
    ) -> None:
        """
················································································

==Endogenize parameters==

················································································
        """
        pass

    @_pages.reference(category="definition", )
    def fix_level(
        self,
        names: Iterable[str] | str,
        /,
    ) -> None:
        """
················································································

==Fix steady levels of variables==

················································································
        """
        pass

    @_pages.reference(category="definition", )
    def fix_change(
        self,
        names: Iterable[str] | str,
        /,
    ) -> None:
        """
················································································

==Fix steady changes of variables==

················································································
        """
        pass


def _resolve_and_check_names(
    register: dict | None,
    names: Iterable[str] | str | EllipsisType,
    /,
) -> tuple[str]:
    """
    """
    keys = tuple(register.keys()) if register else ()
    if names is Ellipsis:
        return keys
    if isinstance(names, str):
        names = (names, )
    names = tuple(names)
    invalid = [n for n in names if n not in keys]
    if invalid:
        raise _wrongdoings.IrisPieError(
            [f"Invalid names:"] + invalid
        )
    return names


def catch_invalid_dates(
    dates: Iterable[Period],
    base_span: tuple[Period],
    /,
) -> NoReturn | None:
    """
    """
    invalid = tuple(repr(d) for d in dates if d not in base_span)
    if invalid:
        raise _wrongdoings.IrisPieError(
            ("These date(s) are out of simulation span:", ) + invalid
        )


PlanSteady = SteadyPlan
Plan = SimulationPlan
PlanSimulate = SimulationPlan




"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, Protocol, runtime_checkable)
from collections.abc import (Iterable, )
import numpy as _np
import wlogging as _wl

from .. import quantities as _quantities
from .. import has_variants as _has_variants
from .. import dates as _dates
from ..databoxes.main import (Databox, )
from ..dataslates.main import (Dataslate, )
from ..fords import solutions as _solutions
from ..fords import simulators as _ford_simulator
from ..periods import simulators as _period_simulator
from ..plans import main as _plans

from . import main as _simultaneous
#]


_SIMULATION_FUNC = {
    "first_order": _ford_simulator.simulate,
    "period": _period_simulator.simulate,
}


InfoOutput = dict[str, Any] | list[dict[str, Any]]


class _SimulationModuleProtocol(Protocol, ):
    """
    """
    #[

    def simulate(self, *args, **kwargs, ) -> dict[str, Any]: ...

    def iter_frames(self, *args, **kwargs, ) -> Iterable[Any]: ...

    #]


class Inlay:
    """
    """
    #[

    def simulate(
        self,
        input_db: Databox,
        span: Iterable[Dater],
        /,
        *,
        plan: _plans.PlanSimulate | None = None,
        method: Literal["first_order", "period", "stacked"] = "first_order",
        prepend_input: bool = True,
        target_databox: Databox | None = None,
        num_variants: int | None = None,
        remove_initial: bool = True,
        remove_terminal: bool = True,
        shocks_from_data: bool = True,
        logging_level: int = _wl.INFO,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> tuple[Databox, InfoOutput]:
        """
        """
        logger = _wl.get_colored_logger(__name__, level=logging_level, )
        num_variants = self.num_variants if num_variants is None else num_variants
        base_dates = tuple(span, )
        #
        extra_databox_names = None
        if plan is not None:
            plan.check_consistency(self, base_dates, )
            extra_databox_names = plan.get_databox_names()
        #
        slatable = self.get_slatable(
            shocks_from_data=shocks_from_data,
        )
        dataslate = Dataslate.from_databox_for_slatable(
            slatable, input_db, base_dates,
            num_variants=num_variants,
            extra_databox_names=extra_databox_names,
        )
        #
        simulation_func = _SIMULATION_FUNC[method]
        zipped = zip(
            range(num_variants, ),
            self.iter_variants(),
            dataslate.iter_variants(),
        )
        #
        #=======================================================================
        # Main loop over variants
        info = []
        for vid, model_v, dataslate_v in zipped:

            info_v = simulation_func(
                model_v, dataslate_v, plan, vid, logger,
                **kwargs,
            )

            info.append(info_v, )

        #=======================================================================
        #
        # Remove initial and terminal condition data (all lags and leads
        # before and after the simulation span)
        if remove_terminal:
            dataslate.remove_terminal()
        if remove_initial:
            dataslate.remove_initial()
        #
        # Convert all variants of the dataslate to a databox
        output_db = dataslate.to_databox()
        if prepend_input:
            output_db.prepend(input_db, base_dates[0]-1, )
        #
        # Add to custom databox
        if target_databox is not None:
            output_db = target_databox | output_db
        #
        is_singleton = num_variants == 1
        info = _has_variants.unpack_singleton(
            info, is_singleton,
            unpack_singleton=unpack_singleton,
        )
        return output_db, info

    #]

